/*
  +----------------------------------------------------------------------+
  | Hidef                                                                |
  +----------------------------------------------------------------------+
  | Copyright (c) 2007 The PHP Group                                     |
  +----------------------------------------------------------------------+
  | This source file is subject to version 3.01 of the PHP license,      |
  | that is bundled with this package in the file LICENSE, and is        |
  | available through the world-wide-web at the following url:           |
  | http://www.php.net/license/3_01.txt                                  |
  | If you did not receive a copy of the PHP license and are unable to   |
  | obtain it through the world-wide-web, please send a note to          |
  | license@php.net so we can mail you a copy immediately.               |
  +----------------------------------------------------------------------+
  | Authors: Gopal Vijayaraghavan <gopalv@php.net>                       |
  +----------------------------------------------------------------------+
*/

/* $Id: hidef.c 311330 2011-05-22 15:40:38Z gopalv $ */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "php.h"
#include "php_ini.h"
#include "php_scandir.h"
#include "zend_globals.h"
#include "zend_ini_scanner.h"
#include "zend_hash.h"
#include "ext/standard/info.h"
#include "SAPI.h"

#include "php_hidef.h"
#include "frozenarray.h"

typedef int (*hidef_walk_dir_cb)(const char* filename, void* ctxt TSRMLS_DC);

/* {{{ stupid stringifcation */
#if DEFAULT_SLASH == '/'
	#define DEFAULT_SLASH_STRING "/"
#elif DEFAULT_SLASH == '\\'
	#define DEFAULT_SLASH_STRING "\\"
#else
	#error "Unknown value for DEFAULT_SLASH"
#endif
/* }}} */


/* {{{ hidef globals 
 *
 * true globals, no need for thread safety here 
 */
static HashTable *hidef_constants_table = NULL;
static HashTable *hidef_data_hash = NULL;
static const char default_ini_path[] = PHP_CONFIG_FILE_SCAN_DIR DEFAULT_SLASH_STRING "hidef";
/* }}} */

/* {{{ PHP_FUNCTION declarations */
PHP_FUNCTION(hidef_fetch);
/* }}} */

/* {{{ ZEND_DECLARE_MODULE_GLOBALS(hidef) */
ZEND_DECLARE_MODULE_GLOBALS(hidef)
/* }}} */

/* {{{ php_hidef_init_globals */
static void php_hidef_init_globals(zend_hidef_globals* hidef_globals TSRMLS_DC)
{
	hidef_globals->ini_path = NULL;
	hidef_globals->data_path = NULL;
}
/* }}} */

/* {{{ php_hidef_shutdown_globals */
static void php_hidef_shutdown_globals(zend_hidef_globals* hidef_globals TSRMLS_DC)
{
	/* nothing ? */
}
/* }}} */

/* {{{ ini entries */
PHP_INI_BEGIN()
STD_PHP_INI_ENTRY("hidef.ini_path",  (char*)default_ini_path,  PHP_INI_SYSTEM, OnUpdateString,       ini_path,   zend_hidef_globals, hidef_globals)
STD_PHP_INI_ENTRY("hidef.data_path", (char*)NULL,              PHP_INI_SYSTEM, OnUpdateString,       data_path,  zend_hidef_globals, hidef_globals)
PHP_INI_END()
/* }}} */

/* {{{ arginfo static macro */
#if PHP_MAJOR_VERSION > 5 || PHP_MAJOR_VERSION == 5 && PHP_MINOR_VERSION >= 3
#define HIDEF_ARGINFO_STATIC
#else
#define HIDEF_ARGINFO_STATIC static
#endif
/* }}} */

/* {{{ arginfo */
HIDEF_ARGINFO_STATIC
ZEND_BEGIN_ARG_INFO(php_hidef_fetch_arginfo, 0)
    ZEND_ARG_INFO(0, key)
    ZEND_ARG_INFO(0, thaw)
ZEND_END_ARG_INFO()
/* }}} */


/* {{{ hidef_functions[]
 *
 * Every user visible function must have an entry in hidef_functions[].
 */
zend_function_entry hidef_functions[] = {
	PHP_FE(hidef_fetch,               php_hidef_fetch_arginfo)
	{NULL, NULL, NULL}	/* Must be the last line in hidef_functions[] */
};
/* }}} */

/* {{{ hidef_module_entry
 */
zend_module_entry hidef_module_entry = {
	STANDARD_MODULE_HEADER,
	"hidef",
	hidef_functions,
	PHP_MINIT(hidef),
	PHP_MSHUTDOWN(hidef),
	NULL,
	NULL,
	PHP_MINFO(hidef),
#if ZEND_MODULE_API_NO >= 20010901
	PHP_HIDEF_VERSION,
#endif
	STANDARD_MODULE_PROPERTIES
};
/* }}} */

#ifdef COMPILE_DL_HIDEF
ZEND_GET_MODULE(hidef)
#endif

/* {{{ struct hidef_parser_ctxt */
typedef struct _hidef_parser_ctxt 
{
	int module_number;
	const char * filename;
	int type;
} hidef_parser_ctxt;
/* }}} */

/* {{{ hidef_zval_pfree */
static void hidef_zval_pfree(void *p) 
{
	TSRMLS_FETCH();
	frozen_array_free_zval_ptr((zval**)p, 1 TSRMLS_CC);
}
/* }}} */

/* {{{ hidef_define_constant */
static int hidef_define_constant(char *key, zval *v, hidef_parser_ctxt* ctxt TSRMLS_DC)
{
	zval value = *v;
	zend_constant c;
	
	c.flags = CONST_CS | CONST_PERSISTENT;
#ifdef CONST_CT_SUBST
	c.flags |= CONST_CT_SUBST;
#endif
	c.module_number = ctxt->module_number;

	zval_copy_ctor(&value); 

	switch(ctxt->type) 
	{
		case IS_LONG:
		{
			convert_to_long(&value);
		}
		break;
		case IS_DOUBLE:
		{
			convert_to_double(&value);
		}
		break;
		case IS_BOOL:
		{
			convert_to_boolean(&value);
		}
		break;
		case IS_STRING:
		{
			convert_to_string(&value);
		}
		break;
	}
			
	/* TODO: check conversion results */
			
	c.value = value;
	if(ctxt->type == IS_STRING) 
	{
		/* TODO: this leaks memory */
		c.value.value.str.val = zend_strndup(Z_STRVAL_P(&c.value), Z_STRLEN_P(&c.value));
	}

	c.name_len = strlen(key)+1;
	c.name = zend_strndup(key, c.name_len-1);
	
	if(zend_register_constant(&c TSRMLS_CC) == FAILURE) 
	{
		php_error_docref(NULL TSRMLS_CC, E_NOTICE, "Constant '%s' redefined in %s line %d\n", key, ctxt->filename, INI_SCNG(lineno) - 1);
		zval_dtor(&value);
		return 0;
	}

	zend_hash_add(hidef_constants_table, c.name, c.name_len, &c.value, sizeof(c.value), NULL);

	zval_dtor(&value);
	return 1;
}
/* }}} */

#ifdef ZEND_BROKEN_INI_SCANNER
/* 
 * php 5.3 changed the ini reader code, which returns two
 * callbacks for "int xyz = 0". one with ("int", NULL, NULL)
 * and then next with ("xyz", "0"). And then changed it back
 * in php 5.3.2+ (in the following revision log)
 *
 * Zend/zend_ini_scanner.l?r1=286913&r2=291525
 *
 * But just to make things "simple", every function is called 
 * hidef_ini_parser_cb().
 */

/* {{{ hidef_ini_parser_cb */
static void hidef_ini_parser_cb(zval *arg1, zval *arg2, zval *arg3, int callback_type, void *arg TSRMLS_DC) 
{
	char *key;
	hidef_parser_ctxt * ctxt = (hidef_parser_ctxt*)arg;

	if(!arg1) return;

	key = Z_STRVAL_P(arg1);
	
	switch (callback_type) 
	{
		case ZEND_INI_PARSER_ENTRY: 
		{
			if(ctxt->type == IS_NULL && !arg2)
			{
				/* callback #1 */	
				if(strncmp(key, "int", strlen("int")) == 0)
				{
					ctxt->type = IS_LONG;
				} 
				else if(strncmp(key, "str", strlen("str")) == 0)
				{
					ctxt->type = IS_STRING;
				} 
				else if(strncmp(key, "float", strlen("float")) == 0)
				{
					ctxt->type = IS_DOUBLE;
				} 
				else if(strncmp(key, "bool", strlen("bool")) == 0)
				{
					ctxt->type = IS_BOOL;
				} 
				else
				{
					ctxt->type = IS_NULL;
				}
			}
			else if(ctxt->type != IS_NULL && arg2)
			{
				hidef_define_constant(key, arg2, ctxt TSRMLS_CC);
				ctxt->type = IS_NULL;
			}
		}
		break;
	
#ifdef ZEND_INI_PARSER_POP_ENTRY
		case ZEND_INI_PARSER_POP_ENTRY: 
		{
			/* do nothing here, I suppose */
		}
		break;
#endif

		case ZEND_INI_PARSER_SECTION:
		{
			/* ignore sectioning */
		}
		break;

		default:
			assert(0);
	}
}
/* }}} */
#else
/* 
 * php 5.2 & php 5.3.x returns one callback with ("int xyz","0"), except the signatures
 * of the callback are different.
 */
/* {{{ hidef_ini_parser_cb */
#ifdef ZEND_ENGINE_2_3
static void hidef_ini_parser_cb(zval *arg1, zval *arg2, zval *arg3, int callback_type, void *arg TSRMLS_DC) 
{
#else
static void hidef_ini_parser_cb(zval *arg1, zval *arg2, int callback_type, void *arg) 
{
	TSRMLS_FETCH();
#endif
	char *p;
	char *key = Z_STRVAL_P(arg1);
	int type = IS_STRING;
	hidef_parser_ctxt * ctxt = (hidef_parser_ctxt*)arg;

	
	switch (callback_type) 
	{
		case ZEND_INI_PARSER_ENTRY: 
		{
			if(!arg2)
			{
				return;
			}

			if(strncmp(key, "int ", strlen("int ")) == 0)
			{
				ctxt->type = IS_LONG;
			} 
			else if(strncmp(key, "str ", strlen("str ")) == 0)
			{
				ctxt->type = IS_STRING;
			} 
			else if(strncmp(key, "float ", strlen("float ")) == 0)
			{
				ctxt->type = IS_DOUBLE;
			} 
			else if(strncmp(key, "bool ", strlen("bool ")) == 0)
			{
				ctxt->type = IS_BOOL;
			} 
			else
			{
				/* TODO: error */
				return;
			}
			
			p = strrchr(key, ' ');

			if((p == NULL) || strlen(p) == 1)
			{
				/* TODO: error handling */
				return;	
			}

			hidef_define_constant(p+1, arg2, ctxt TSRMLS_CC);

		}
		break;
	
#ifdef ZEND_INI_PARSER_POP_ENTRY
		case ZEND_INI_PARSER_POP_ENTRY: 
		{
			/* do nothing here, I suppose */
		}
		break;
#endif

		case ZEND_INI_PARSER_SECTION:
		{
			/* ignore sectioning */
		}
		break;

		default:
			assert(0);
	}
}
/* }}} */
#endif

/* {{{ hidef_walk_dir */
static int hidef_walk_dir(const char *path, const char *ext, hidef_walk_dir_cb cb, void *ctxt TSRMLS_DC)
{
	char file[MAXPATHLEN]={0,};
	int ndir, i, k;
	char *p = NULL;
	struct dirent **namelist = NULL;

	if ((ndir = php_scandir(path, &namelist, 0, php_alphasort)) > 0)
	{
		for (i = 0; i < ndir; i++) 
		{
			/* check for extension */
			if (!(p = strrchr(namelist[i]->d_name, '.')) 
					|| (p && strcmp(p, ext))) 
			{
				free(namelist[i]);
				continue;
			}
			snprintf(file, MAXPATHLEN, "%s%c%s", 
					path, DEFAULT_SLASH, namelist[i]->d_name);
			if(!cb(file, ctxt TSRMLS_CC))
			{
				goto cleanup;
			}
			free(namelist[i]);
		}
		free(namelist);
	}

	return 1;

cleanup:
	for(k = i; k < ndir; k++)
	{
		free(namelist[k]);
	}
	free(namelist);

	return 0;
}
/* }}} */

/* {{{ hidef_parse_ini */
static int hidef_parse_ini(const char *ini_file, void* pctxt TSRMLS_DC)
{
	struct stat sb;
	zend_file_handle fh = {0,};
	hidef_parser_ctxt *ctxt = (hidef_parser_ctxt*)pctxt;

	if (VCWD_STAT(ini_file, &sb) == 0 && S_ISREG(sb.st_mode)) 
	{
		if ((fh.handle.fp = VCWD_FOPEN(ini_file, "r"))) 
		{
			fh.filename = (char*)ini_file;
			ctxt->filename = ini_file;
			fh.type = ZEND_HANDLE_FP;
#ifdef ZEND_ENGINE_2_3
			zend_parse_ini_file(&fh, 1, ZEND_INI_SCANNER_RAW, (zend_ini_parser_cb_t) hidef_ini_parser_cb, ctxt TSRMLS_CC);
#else
			zend_parse_ini_file(&fh, 1, hidef_ini_parser_cb, ctxt);
#endif
			return 1;
		}
	}

	return 0;
}
/* }}} */

/* {{{ hidef_read_ini */
static int hidef_read_ini(hidef_parser_ctxt *ctxt TSRMLS_DC)
{
	const char *ini_path = NULL;

	if(HIDEF_G(ini_path)) 
	{
		ini_path = HIDEF_G(ini_path);
	}
	else if(sizeof(PHP_CONFIG_FILE_SCAN_DIR) > 1) 
	{
		ini_path = (const char *)default_ini_path;
	}
	else
	{
		return 0;
	}

	return hidef_walk_dir(ini_path, ".ini", hidef_parse_ini, ctxt TSRMLS_CC);
}
/* }}} */

/* {{{ hidef_load_data */
static int hidef_load_data(const char *data_file, void* pctxt TSRMLS_DC)
{
	char *p;
	char key[MAXPATHLEN] = {0,};
	unsigned int key_len;
	zval *data;

	p = strrchr(data_file, DEFAULT_SLASH);

	if(p && p[1])
	{
		strlcpy(key, p+1, sizeof(key));
		p = strrchr(key, '.');

		if(p)
		{
			p[0] = '\0';
			key_len = strlen(key);

			zend_try 
			{
				data = frozen_array_unserialize(data_file TSRMLS_CC);
			} zend_catch
			{
				zend_error(E_ERROR, "Data corruption in %s, bailing out", data_file);
				zend_bailout();
			} zend_end_try();

			if((data == NULL) || zend_hash_add(hidef_data_hash, key, key_len+1, &data, sizeof(void*), NULL) == FAILURE)
			{
				zend_error(E_ERROR, "Unable to add %s to the hidef data hash", data_file);
				return 0;
			}
			return 1;
		}
	}

	return 0;
}
/* }}} */

/* {{{ hidef_read_data */
static int hidef_read_data(hidef_parser_ctxt *ctxt TSRMLS_DC)
{
	const char *data_path = NULL;

	if(HIDEF_G(data_path)) 
	{
		data_path = HIDEF_G(data_path);
	}
	else
	{
		return 0;
	}

	return hidef_walk_dir(data_path, ".data", hidef_load_data, ctxt TSRMLS_CC);
}
/* }}} */

/* {{{ PHP_MINIT_FUNCTION
 */
PHP_MINIT_FUNCTION(hidef)
{
	hidef_parser_ctxt ctxt = {0,};

	ZEND_INIT_MODULE_GLOBALS(hidef, php_hidef_init_globals, php_hidef_shutdown_globals);

	REGISTER_INI_ENTRIES();
	
	ctxt.module_number = module_number;

	hidef_constants_table = pemalloc(sizeof(HashTable),1);
	hidef_data_hash = pemalloc(sizeof(HashTable),1);

	zend_hash_init(hidef_constants_table, 32,  NULL, NULL, 1);
	zend_hash_init(hidef_data_hash, 	  32,  NULL, hidef_zval_pfree, 1);

	hidef_read_ini(&ctxt TSRMLS_CC);
	hidef_read_data(&ctxt TSRMLS_CC);
	
	frozen_array_init(TSRMLS_C);

	return SUCCESS;
}
/* }}} */

/* {{{ PHP_MSHUTDOWN_FUNCTION
 *  */
PHP_MSHUTDOWN_FUNCTION(hidef)
{
	zend_hash_destroy(hidef_constants_table);
	zend_hash_destroy(hidef_data_hash);

	pefree(hidef_constants_table, 1);
	pefree(hidef_data_hash, 1);

#ifdef ZTS
	ts_free_id(hidef_globals_id);
#else
	php_hidef_shutdown_globals(&hidef_globals);
#endif

	UNREGISTER_INI_ENTRIES();
	return SUCCESS;
}
/* }}} */

/* {{{ PHP_MINFO_FUNCTION
 */
PHP_MINFO_FUNCTION(hidef)
{
	const char * ini_path = NULL;
	const char * data_path = "disabled";
	HashPosition pos;
	zend_constant *val;
	int module_number = zend_module->module_number;
	
	php_info_print_table_start();

	if(HIDEF_G(ini_path)) 
	{
		ini_path = HIDEF_G(ini_path);	
	}
	else if(sizeof(PHP_CONFIG_FILE_SCAN_DIR) > 1)
	{
		ini_path = (const char *)default_ini_path;
	}
	else
	{
		php_info_print_table_header(2, "hidef support", "disabled");
		php_info_print_table_end();
		return;
	}
	
	if(HIDEF_G(data_path)) 
	{
		data_path = HIDEF_G(data_path);	
	}
	
	
	php_info_print_table_header(2, "hidef support", "enabled");
	php_info_print_table_row(2, "version", PHP_HIDEF_VERSION);
	php_info_print_table_row(2, "ini search path", ini_path);
	php_info_print_table_row(2, "data search path", data_path);
#ifdef CONST_CT_SUBST
	php_info_print_table_row(2, "substitution mode", "compile time");
#else
	php_info_print_table_row(2, "substitution mode", "runtime");
#endif

	php_info_print_table_end();
	php_info_print_table_start();
	php_info_print_table_header(2, "Constant", "Value");

	
	zend_hash_internal_pointer_reset_ex(EG(zend_constants), &pos);
	while (zend_hash_get_current_data_ex(EG(zend_constants), (void **) &val, &pos) != FAILURE) 
	{
		if (val->module_number == module_number) 
		{
			zval const_val = {{0,},};
			const_val = val->value;
			zval_copy_ctor(&const_val);
			convert_to_string(&const_val);
			
			php_info_print_table_row(2, val->name, Z_STRVAL_P(&const_val));
			zval_dtor(&const_val);
		}
		zend_hash_move_forward_ex(EG(zend_constants), &pos);
	}

	php_info_print_table_end();
}
/* }}} */

/* {{{ proto mixed hidef_fetch(string key [, bool thaw])
 */
PHP_FUNCTION(hidef_fetch) 
{
	zval **hentry;
	zval *wrapped;
	char *strkey;
	int strkey_len;
	zend_bool thaw = 0;

	if(zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "s|b", &strkey, &strkey_len, &thaw) == FAILURE)
	{
		return;
	}

	if(zend_hash_find(hidef_data_hash, strkey, strkey_len+1, (void**)&hentry) == FAILURE)
	{
		return;
	}

	if(thaw)
	{
		wrapped = frozen_array_copy_zval_ptr(NULL, hentry[0], 0, NULL TSRMLS_CC);
	}
	else
	{
		wrapped = frozen_array_wrap_zval(hentry[0] TSRMLS_CC);
	}

	RETURN_ZVAL(wrapped, 0, 1);
}
/* }}} */

/*
 * Local variables:
 * tab-width: 4
 * c-basic-offset: 4
 * End:
 * vim>600: noet sw=4 ts=4 fdm=marker
 * vim<600: noet sw=4 ts=4
 */
